unit XForms;

interface

uses Windows, Messages, XClasses, XWindows, XVisuals, XGraphics;

const
  CN_BASE = $BC00;
  CN_CHARTOITEM        = CN_BASE + WM_CHARTOITEM;
  CN_COMMAND           = CN_BASE + WM_COMMAND;
  CN_COMPAREITEM       = CN_BASE + WM_COMPAREITEM;
  CN_CTLCOLORBTN       = CN_BASE + WM_CTLCOLORBTN;
  CN_CTLCOLORDLG       = CN_BASE + WM_CTLCOLORDLG;
  CN_CTLCOLOREDIT      = CN_BASE + WM_CTLCOLOREDIT;
  CN_CTLCOLORLISTBOX   = CN_BASE + WM_CTLCOLORLISTBOX;
  CN_CTLCOLORMSGBOX    = CN_BASE + WM_CTLCOLORMSGBOX;
  CN_CTLCOLORSCROLLBAR = CN_BASE + WM_CTLCOLORSCROLLBAR;
  CN_CTLCOLORSTATIC    = CN_BASE + WM_CTLCOLORSTATIC;
  CN_DELETEITEM        = CN_BASE + WM_DELETEITEM;
  CN_DRAWITEM          = CN_BASE + WM_DRAWITEM;
  CN_HSCROLL           = CN_BASE + WM_HSCROLL;
  CN_MEASUREITEM       = CN_BASE + WM_MEASUREITEM;
  CN_PARENTNOTIFY      = CN_BASE + WM_PARENTNOTIFY;
  CN_VKEYTOITEM        = CN_BASE + WM_VKEYTOITEM;
  CN_VSCROLL           = CN_BASE + WM_VSCROLL;
  CN_KEYDOWN           = CN_BASE + WM_KEYDOWN;
  CN_KEYUP             = CN_BASE + WM_KEYUP;
  CN_CHAR              = CN_BASE + WM_CHAR;
  CN_SYSKEYDOWN        = CN_BASE + WM_SYSKEYDOWN;
  CN_SYSCHAR           = CN_BASE + WM_SYSCHAR;
  CN_NOTIFY            = CN_BASE + WM_NOTIFY;

type
   XAlign = ( alNone, alLeft, alRight, alTop, alBottom, alClient );
   XTextAlignment = (taLeft, taCenter, taRight);

   XMouseButton = ( mbLeft, mbRight, mbMiddle );
   XShiftButton = ( sbLeft, sbRight, sbMiddle, sbShift, sbCtrl, sbAlt );
   XShiftState = Set of XShiftButton;
   XOnMouse = procedure( Sender : TObject; Shift : XShiftState; X, Y : Integer )
              of object;
   XOnMouseUpDown = procedure( Sender : TObject; Button : XMouseButton;
                               Shift : XShiftState; X, Y : Integer ) of object;
   XOnKey = procedure( Sender : TObject; var Key : Word; Shift : XShiftState ) of object;
   XOnKeyPress = procedure( Sender : TObject; var Key : Char ) of object;

   XControl = class;

   XOnFormMessage = procedure( Sender : TObject; var Msg : TMsg; var Rslt : Integer;
                               var Handled : Boolean ) of object;

   XForm = class( XWindow )
   private
      fCurrentControl : XControl;
      fOnMouseMove : XOnMouse;
      fOnMouseDown : XOnMouseUpDown;
      fOnMouseUp : XOnMouseUpDown;
      fOnResize : XOnEvent;
      fAutoSize : Boolean;
      fMargin : Integer;
      FOnKeyUp: XOnKey;
      FOnKeyDown: XOnKey;
      FOnKeyPress: XOnKeyPress;
      fCtrlCreateIdx : Integer;
      fUpdRgn : HRgn;
    FOnMessage: XOnFormMessage;
      procedure SetCurrentControl( Value : XControl );
      procedure SetAutoSize( Value : Boolean );
      procedure SetMargin( Value : Integer );
   protected
      procedure Paint; dynamic;
      {
      function KeyDown( var Msg : TMsg ) : Boolean;
      function KeyUp( var Msg : TMsg ) : Boolean;
      function KeyChar( var Msg : TMsg ) : Boolean;
      }
      function GetWndUpdRgn : HRgn; override;
      procedure ExcludeUpdRect( DC : HDC; Rect : TRect ); override;
      procedure ExcludeUpdRgn( DC : HDC; Rgn : HRgn ); override;
   public
      constructor Create( AParent : XClass ); override;
      property CurrentControl : XControl read fCurrentControl write SetCurrentControl;
      property On_Message : XOnFormMessage read FOnMessage write FOnMessage;
      property OnMouseMove : XOnMouse read fOnMouseMove write fOnMouseMove;
      property OnMouseDown : XOnMouseUpDown read fOnMouseDown write fOnMouseDown;
      property OnMouseUp : XOnMouseUpDown read fOnMouseUp write fOnMouseUp;
      property OnKeyDown : XOnKey read FOnKeyDown write FOnKeyDown;
      property OnKeyUp : XOnKey read FOnKeyUp write FOnKeyUp;
      property OnKeyPress : XOnKeyPress read FOnKeyPress write FOnKeyPress;
      function AvailableClientRect : TRect;
      property OnResize : XOnEvent read fOnResize write fOnResize;
      procedure GotoControl( VKey : Word );
      property AutoSize : Boolean read fAutoSize write SetAutoSize;
      property Margin : Integer read fMargin write SetMargin;
      function ControlAtPos( X, Y : Integer; IgnoreDisabled : Boolean ) : XControl;
      function CreateWindow : Boolean; override;
      function WndProc( var Msg : TMsg ) : Integer; override;
   end;

   XPlacePos = ( ppRight, ppLeft, ppBottom, ppTop );

   XControl = class( XVisual )
   private
      fOwnerForm : XForm;
      fAlign : XAlign;
      fOnMouseMove : XOnMouse;
      fOnMouseDown : XOnMouseUpDown;
      fOnMouseUp : XOnMouseUpDown;
    FOnKeyUp: XOnKey;
    FOnKeyDown: XOnKey;
    FOnKeyPress: XOnKeyPress;
    fTabOrder: Integer;
      function GetShowing : Boolean;
      procedure SetAlign( Value : XAlign );
      function GetParentForm : XForm;
      function GetFocused : Boolean;
   protected
      procedure SetOwnerForm( Value : XForm ); virtual;
      function GetWndUpdRgn : HRgn; override;
      function GetEnabled : Boolean; override;
      procedure SetEnabled( Value : Boolean ); override;
      procedure SetVisible( Value : Boolean ); override;
      procedure SetBoundsRect( Value : TRect ); override;
      procedure SetFocused( Value : Boolean ); virtual;
   public
      procedure DoMouseMove( Shift : XShiftState; X, Y : Integer ); dynamic;
      procedure DoMouseDown( Button : XMouseButton; Shift : XShiftState; X, Y : Integer ); dynamic;
      procedure DoMouseUp( Button : XMouseButton; Shift : XShiftState; X, Y : Integer ); dynamic;
      procedure DoKeyDown( var Key : Word; Shift : XShiftState ); dynamic;
      procedure DoKeyUp( var Key : Word; Shift : XShiftState ); dynamic;
      procedure DoKeyPress( var Key : Char ); dynamic;
      constructor Create( AParent : XClass ); override;
      destructor Destroy; override;
      function CanFocus : Boolean; dynamic; abstract;
      property Focused : Boolean read GetFocused write SetFocused;
      property OwnerForm : XForm read fOwnerForm write SetOwnerForm;
      property Visible;
      property Showing : Boolean read GetShowing;
      property Enabled;
      property Align : XAlign read fAlign write SetAlign;
      property ParentForm : XForm read GetParentForm;
      property Left;
      property Top;
      property Width;
      property Height;
      property BoundsRect;
      function ClientToScreen( P : TPoint ) : TPoint;
      function ScreenToClient( P : TPoint ) : TPoint;
      property OnMouseMove : XOnMouse read fOnMouseMove write fOnMouseMove;
      property OnMouseDown : XOnMouseUpDown read fOnMouseDown write fOnMouseDown;
      property OnMouseUp : XOnMouseUpDown read fOnMouseUp write fOnMouseUp;
      property OnKeyDown : XOnKey read FOnKeyDown write FOnKeyDown;
      property OnKeyUp : XOnKey read FOnKeyUp write FOnKeyUp;
      property OnKeyPress : XOnKeyPress read FOnKeyPress write FOnKeyPress;
      function ClientRect : TRect; override;
      function AvailableClientRect : TRect;
      function ControlAtPos( X, Y : Integer; IgnoreDisabled : Boolean ) : XControl;
      procedure Invalidate; override;
      procedure Place( Control : XControl; NewPos : XPlacePos; Align : Boolean );
      property TabOrder : Integer read fTabOrder;
      procedure ExcludeUpdRect( DC : HDC; Rect : TRect ); override;
      procedure ExcludeUpdRgn( DC : HDC; Rgn : HRgn ); override;
      function Windowed : Boolean; override;
   end;

function GetClassParentForm( C : XClass ) : XForm;

type
   XOnDrawFocusRect = procedure( Sender : TObject; Rect : TRect; Color : XColor ) of object;
   XOnGotoControl = procedure(Sender : TObject; VKey : DWord ) of object;
   XOnGetAvailableClientRect = procedure( Sender : TObject; var Rect : TRect )
                               of object;

   XFormsManager = class
   private
    FOnDrawFocusRect: XOnDrawFocusRect;
    FOnFormMessage : XOnFormMessage;
    FOnGotoControl : XOnGotoControl;
      fOnGetAvailableClientRect : XOnGetAvailableClientRect;
      fOnAutoSizeForm : XOnEvent;
      fOnRealign : XOnEvent;
   protected
   public
      property OnDrawFocusRect : XOnDrawFocusRect read FOnDrawFocusRect write fOnDrawFocusRect;
      property On_FormMessage : XOnFormMessage read FOnFormMessage write FOnFormMessage;
      property OnGotoControl : XOnGotoControl read FOnGotoControl write FOnGotoControl;
      property On_AutoSizeForm : XOnEvent read fOnAutoSizeForm write fOnAutoSizeForm;
      property On_GetAvailableClientRect : XOnGetAvailableClientRect
                                        read fOnGetAvailableClientRect
                                        write fOnGetAvailableClientRect;
      property On_Realign : XOnEvent read fOnRealign write fOnRealign;
   end;

var GlobalFormsManager : XFormsManager;

implementation

{ XControl implementation }

constructor XControl.Create( AParent : XClass );
var F : XForm;
    C : XClass;
begin
   inherited;
   Visible := True;
   if AParent is XForm then
      fOwnerForm := AParent as XForm
   else
   if AParent is XControl then
      fOwnerForm := (AParent as XControl).OwnerForm;
   if CanFocus then
   begin
      F := ParentForm;
      if F <> nil then
      begin
         Inc( F.fCtrlCreateIdx );
         fTabOrder := F.fCtrlCreateIdx;
         C := Parent;
         while C <> nil do
         begin
            if C is XControl then
            begin
               (C as XControl).fTabOrder := 0;
               C := C.Parent;
            end
               else
            if C is XForm then
               break;
         end;
      end;
   end;
end;

destructor XControl.Destroy;
begin
   Visible := False;
   inherited;
end;

function XControl.GetEnabled : Boolean;
begin
   Result := inherited GetEnabled;
   if Parent = nil then Exit;
   if Parent is XVisual then
      Result := Result and (Parent as XVisual).Enabled;
end;

procedure XControl.SetEnabled( Value : Boolean );
begin
   if Enabled = Value then Exit;
   inherited SetEnabled( Value );
   if Enabled then
   begin
      if ParentForm.CurrentControl = nil then
         ParentForm.CurrentControl := Self;
   end
      else
   begin
      if ParentForm.CurrentControl = Self then
         ParentForm.GotoControl( VK_TAB );
   end;
end;

procedure XControl.SetVisible( Value : Boolean );
var WasShow, NowShow : Boolean;
    F : XForm;
begin
   if Visible = Value then Exit;
   WasShow := Showing;
   inherited SetVisible( Value );
   NowShow := Showing;
   F := ParentForm;
   if F <> nil then
   if NowShow <> WasShow then
   begin
      if not NowShow then
      if F.CurrentControl = Self then
         F.CurrentControl := nil;
      F.Invalidate;
   end;
end;

function XControl.GetShowing : Boolean;
var V : XVisual;
begin
   Result := Visible;
   V := Self;
   while Result do
   begin
      if V.Parent = nil then break;
      V := V.Parent as XVisual;
      Result := V.Visible;
      if V is XForm then break;
   end;
end;

procedure XControl.SetAlign( Value : XAlign );
var R : TRect;
    F : XForm;
begin
   if fAlign = Value then Exit;
   if Value <> alNone then
   if Parent <> nil then
   begin
      if Parent is XControl then
         R := (Parent as XControl).AvailableClientRect
      else
      if Parent is XForm then
         R := (Parent as XForm).AvailableClientRect
      else Exit;
      case Value of
      alLeft: R.Right := R.Left + Width;
      alTop : R.Bottom := R.Top + Height;
      alRight: begin
                 if R.Right < Width then
                    R.Right := Width;
                 R.Left := R.Right - Width;
               end;
      alBottom:begin
                 if R.Bottom > Height then
                    R.Bottom := Height;   
                 R.Top := R.Bottom - Height;
               end;
      end;
      BoundsRect := R;
   end;
   fAlign := Value;

   F := ParentForm;
   if F <> nil then
   if Assigned( GlobalFormsManager.fOnRealign ) then
      GlobalFormsManager.fOnRealign( Parent );
end;

function GetClassParentForm( C : XClass ) : XForm;
begin
   C := C.Parent;
   while C <> nil do
   begin
      if C is XForm then
      begin
         Result := C as XForm;
         Exit;
      end;
      C := C.Parent;
   end;
   Result := nil;
end;

function XControl.GetParentForm : XForm;
begin
   Result := GetClassParentForm( Self );
end;

procedure XControl.SetBoundsRect( Value : TRect );
var R : TRect;
    F : XForm;
    Show : Boolean;
begin
   R := BoundsRect;
   if (R.Left <> Value.Left) or (R.Top <> Value.Top)
   or (R.Right <> Value.Right) or (R.Bottom <> Value.Bottom) then
   begin
     Show := Showing;
     if Show then
        Invalidate;
     inherited SetBoundsRect( Value );
     if Show then
        Invalidate;
   end;
   F := ParentForm;
   if F <> nil then
   begin
      if F.AutoSize then
      begin
        if Assigned( GlobalFormsManager.fOnAutoSizeForm ) then
           GlobalFormsManager.fOnAutoSizeForm( F );
      end;
   end;
end;

function XControl.ClientToScreen( P : TPoint ) : TPoint;
begin
   Result.x := P.x + Left;
   Result.y := P.y + Top;
   if Parent is XControl then
   begin
      Result := (Parent as XControl).ClientToScreen( Result );
      Exit;
   end;
   Result := ParentForm.ClientToScreen( Result );
end;

function XControl.ScreenToClient( P : TPoint ) : TPoint;
begin
   Result := P;
   if Parent is XControl then
   begin
      Result := (Parent as XControl).ScreenToClient( Result );
   end
      else
   begin
      Result := ParentForm.ScreenToClient( Result );
   end;
   Result.x := Result.x - Left;
   Result.y := Result.y - Top;
end;

function XControl.Windowed : Boolean;
begin
   Result := False;
end;

procedure XControl.DoMouseMove( Shift : XShiftState; X, Y : Integer );
var C : XControl;
begin
   C := ControlAtPos( X, Y, True );
   if C <> nil then
      C.DoMouseMove( Shift, X - C.Left, Y - C.Top )
   else
   if Assigned( fOnMouseMove ) then
      fOnMouseMove( Self, Shift, X, Y );
end;

procedure XControl.DoMouseDown( Button : XMouseButton; Shift : XShiftState; X, Y : Integer );
var C : XControl;
    R : TRect;
begin
   C := ControlAtPos( X, Y, True );
   if C <> nil then
   begin
      R := ClientRect;
      C.DoMouseDown( Button, Shift, X - C.Left - R.Left, Y - C.Top - R.Top );
   end
      else
   begin
      Focused := True;
      if Assigned( fOnMouseDown ) then
         fOnMouseDown( Self, Button, Shift, X, Y );
   end;
end;

procedure XControl.DoMouseUp( Button : XMouseButton; Shift : XShiftState; X, Y : Integer );
var C : XControl;
begin
   C := ControlAtPos( X, Y, True );
   if C <> nil then
      C.DoMouseUp( Button, Shift, X - C.Left, Y - C.Top )
   else
   if Assigned( fOnMouseUp ) then
      fOnMouseUp( Self, Button, Shift, X, Y );
end;

function XControl.AvailableClientRect : TRect;
begin
   Result := ClientRect;
   if ParentForm <> nil then
   if Assigned( GlobalFormsManager.fOnGetAvailableClientRect ) then
      GlobalFormsManager.fOnGetAvailableClientRect( Self, Result );
end;

function XControl.ClientRect : TRect;
begin
   Result := XRect( 0, 0, Width, Height );
end;

function XControl.ControlAtPos( X, Y : Integer; IgnoreDisabled : Boolean )
                             : XControl;
var V : XVisual;
begin
   Result := nil;
   V := ElementAtPos( X, Y, IgnoreDisabled );
   if V = nil then Exit;
   if not (V is XControl) then Exit;
   Result := V as XControl;
end;

function XControl.GetWndUpdRgn: HRgn;
begin
   Result := ParentForm.fUpdRgn;
end;

procedure XControl.Invalidate;
var F : XForm;
    R : TRect;
begin
   if not Showing then Exit;
   F := ParentForm;
   if F = nil then Exit;
   if F.Handle = 0 then Exit;
   R := ControlRect;
   InvalidateRect( F.Handle, @R, False );
end;

procedure XControl.Place( Control : XControl; NewPos : XPlacePos; Align : Boolean );
var R, B : TRect;
    M : Integer;
    F : XForm;
begin
   if Control.Parent <> Parent then Exit;
   R := Control.BoundsRect;
   B := BoundsRect;
   M := 4;
   F := ParentForm;
   if F <> nil then M := F.Margin;
   case NewPos of
   ppLeft: begin
             OffsetRect( B, -(B.Left - R.Left) - M - (B.Right - B.Left), 0 );
             if Align then
                OffsetRect( B, 0, -(B.Top - R.Top) );
           end;
   ppRight:
           begin
             OffsetRect( B, -(B.Left - R.Left) + (R.Right - R.Left) + M, 0 );
             if Align then
                OffsetRect( B, 0, -(B.Top - R.Top) );
           end;
   ppTop:  begin
             OffsetRect( B, 0, -(B.Top - R.Top) - M - (B.Bottom - B.Top) );
             if Align then
                OffsetRect( B, -(B.Left - R.Left), 0 );
           end;
   ppBottom:
           begin
             OffsetRect( B, 0, -(B.Top - R.Top) + (R.Bottom - R.Top) + M );
             if Align then
                OffsetRect( B, -(B.Left - R.Left), 0 );
           end;
   end;
   BoundsRect := B;
end;

procedure XControl.SetOwnerForm( Value : XForm );
begin
   fOwnerForm := Value;
end;

function XControl.GetFocused : Boolean;
var F : XForm;
begin
   Result := False;
   F := ParentForm;
   if F = nil then Exit;
   Result := F.CurrentControl = Self;
end;

procedure XControl.SetFocused( Value : Boolean );
var F : XForm;
    I : Integer;
    C : XClass;
    V : XControl;
begin
   if not CanFocus then Exit;
   F := ParentForm;
   if F = nil then Exit;
   if Value then
   begin
      for I := 0 to ChildsCount - 1 do
      begin
         C := Childs[ I ];
         if C is XControl then
         begin
            V := C as XControl;
            if V.CanFocus then
            begin
               V.SetFocused( True );
               Exit;
            end;
         end;
      end;
      F.CurrentControl := Self;
   end
      else
   if Focused then
      F.GotoControl( VK_TAB );
end;

procedure XControl.DoKeyDown(var Key: Word; Shift: XShiftState);
begin
   if Assigned( fOnKeyDown ) then
      fOnKeyDown( Self, Key, Shift );
end;

procedure XControl.DoKeyPress(var Key: Char);
begin
   if Assigned( fOnKeyPress ) then
      fOnKeyPress( Self, Key );
end;

procedure XControl.DoKeyUp(var Key: Word; Shift: XShiftState);
begin
   if Assigned( fOnKeyUp ) then
      fOnKeyUp( Self, Key, Shift );
end;

procedure XControl.ExcludeUpdRect(DC: HDC; Rect: TRect);
begin
   if ParentForm = nil then inherited
   else ParentForm.ExcludeUpdRect( DC, Rect );
end;

procedure XControl.ExcludeUpdRgn( DC : HDC; Rgn : HRgn );
begin
   if ParentForm = nil then inherited
   else ParentForm.ExcludeUpdRgn( DC, Rgn );
end;

{ XForm implementation }

constructor XForm.Create( AParent : XClass );
begin
   inherited;
   fMargin := 4;
   fAutoSize := True;
end;

procedure XForm.SetCurrentControl( Value : XControl );
   procedure InvalidateCurrent;
   begin
      if fCurrentControl <> nil then
         fCurrentControl.Invalidate;
   end;
begin
   if Value = CurrentControl then Exit;
   if Value <> nil then
   begin
      if Value.ParentForm <> Self then Exit;
      if not Value.Showing then Exit;
      if not Value.CanFocus then Exit;
   end;
   InvalidateCurrent;
   fCurrentControl := nil;
   UpdateWindow( Handle );
   fCurrentControl := Value;
   InvalidateCurrent;
end;

function XForm.WndProc( var Msg : TMsg ) : Integer;
var {Shift : XShiftState;
    X, Y : Integer;
    C : XControl;
    }
    Handled : Boolean;
begin
   Result := 0;
   Handled := False;
   if Assigned( GlobalFormsManager.On_FormMessage ) then
   begin
      GlobalFormsManager.On_FormMessage( Self, Msg, Result, Handled );
      if Handled then Exit;
   end;
   if Assigned( On_Message ) then
   begin
      On_Message( Self, Msg, Result, Handled );
      if Handled then Exit;
   end;
   case Msg.message of
   WM_SIZE :begin
               Result := inherited WndProc( Msg );
               if Assigned( GlobalFormsManager.fOnRealign ) then
                  GlobalFormsManager.fOnRealign( Self );
               if Assigned( fOnResize ) then
                  fOnResize( Self );
            end;
   WM_PAINT:  begin Paint; Result := 0; end;

   WM_ERASEBKGND: begin
                    Result := 1;
                    Exit;
                  end;
   // Introduced here, because these messages are sent to parent
   // window of message box, scrollbar and static MFC control,
   // not to window itself:
   WM_CTLCOLORMSGBOX..WM_CTLCOLORSTATIC:
       Result := SendMessage(Msg.LParam, CN_BASE + Msg.message, Msg.WParam, Msg.LParam);
   else Result := inherited WndProc( Msg );
   end;
end;

function XForm.AvailableClientRect : TRect;
begin
   Result := ClientRect;
   if Assigned( GlobalFormsManager.fOnGetAvailableClientRect ) then
      GlobalFormsManager.fOnGetAvailableClientRect( Self, Result );
end;

procedure XForm.GotoControl( VKey : Word );
begin
   if Assigned( GlobalFormsManager.OnGotoControl ) then
      GlobalFormsManager.OnGotoControl( Self, VKey );
end;

procedure XForm.SetAutoSize( Value : Boolean );
begin
   if fAutoSize = Value then Exit;
   fAutoSize := Value;
   if Value then
   if Assigned( GlobalFormsManager.fOnAutoSizeForm ) then
      GlobalFormsManager.fOnAutoSizeForm( Self );
end;

procedure XForm.SetMargin( Value : Integer );
begin
   if fMargin = Value then Exit;
   fMargin := Value;
   if AutoSize then
   if Assigned( GlobalFormsManager.fOnAutoSizeForm ) then
      GlobalFormsManager.fOnAutoSizeForm( Self );
end;

function XForm.CreateWindow : Boolean;
begin
   Result := inherited CreateWindow;
   if AutoSize then
   if Assigned( GlobalFormsManager.fOnAutoSizeForm ) then
      GlobalFormsManager.fOnAutoSizeForm( Self );
   GotoControl( VK_Home );
end;

function XForm.ControlAtPos( X, Y : Integer; IgnoreDisabled : Boolean )
                             : XControl;
var V : XVisual;
begin
   Result := nil;
   V := ElementAtPos( X, Y, IgnoreDisabled );
   if V = nil then Exit;
   if not (V is XControl) then Exit;
   Result := V as XControl;
end;

procedure XForm.Paint;
var
    UpdCmplx : Integer;
    FillBrush : HBrush;
    DC : HDC;
    Rect : TRect;
begin
   if Handle = 0 then Exit;
   fUpdRgn := CreateRectRgn( 0, 0, 0, 0 );
   UpdCmplx := GetUpdateRgn( Handle, fUpdRgn, False );
   if UpdCmplx = ERROR then Exit;
   if UpdCmplx <> NULLREGION then
   begin
      DC := GetDC( Handle );
      SelectClipRgn( DC, fUpdRgn );
      if Assigned( GlobalWindowsManager.OnBeginPaint ) then
         GlobalWindowsManager.OnBeginPaint( Self, DC );
      GetRgnBox( fUpdRgn, Rect );
      DoPaint( DC, Rect );
         FillBrush := CreateSolidBrush( Color2RGB( Color ) );
         Windows.FillRgn( DC, fUpdRgn, FillBrush );
         DeleteObject( FillBrush );
      if Assigned( GlobalWindowsManager.OnEndPaint ) then
         GlobalWindowsManager.OnEndPaint( Self, DC );
      ReleaseDC( Handle, DC );
   end;
   ValidateRect( Handle, nil );
   DeleteObject( fUpdRgn ); fUpdRgn := 0;
end;

function XForm.GetWndUpdRgn: HRgn;
begin
   Result := fUpdRgn;
end;

procedure XForm.ExcludeUpdRect(DC: HDC; Rect: TRect);
var Rgn : HRgn;
begin
   if fUpdRgn <> 0 then
   begin
      Rgn := CreateRectRgn( Rect.Left, Rect.Top, Rect.Right, Rect.Bottom );
      CombineRgn( fUpdRgn, fUpdRgn, Rgn, RGN_DIFF );
      DeleteObject( Rgn );
      SelectClipRgn( DC, fUpdRgn );
   end
      else
   inherited;
end;

procedure XForm.ExcludeUpdRgn(DC: HDC; Rgn: HRgn);
begin
   if fUpdRgn <> 0 then
   begin
      CombineRgn( fUpdRgn, fUpdRgn, Rgn, RGN_DIFF );
      SelectClipRgn( DC, fUpdRgn );
   end
      else
   inherited;
end;

initialization

   GlobalFormsManager := XFormsManager.Create;

finalization

   GlobalFormsManager.Free;

end.
