unit XGraphics;

interface

uses Windows, XClasses, XStrUtils;

type
   XColor = Integer;

const
  clScrollBar = XColor(COLOR_SCROLLBAR or $80000000);
  clBackground = XColor(COLOR_BACKGROUND or $80000000);
  clActiveCaption = XColor(COLOR_ACTIVECAPTION or $80000000);
  clInactiveCaption = XColor(COLOR_INACTIVECAPTION or $80000000);
  clMenu = XColor(COLOR_MENU or $80000000);
  clWindow = XColor(COLOR_WINDOW or $80000000);
  clWindowFrame = XColor(COLOR_WINDOWFRAME or $80000000);
  clMenuText = XColor(COLOR_MENUTEXT or $80000000);
  clWindowText = XColor(COLOR_WINDOWTEXT or $80000000);
  clCaptionText = XColor(COLOR_CAPTIONTEXT or $80000000);
  clActiveBorder = XColor(COLOR_ACTIVEBORDER or $80000000);
  clInactiveBorder = XColor(COLOR_INACTIVEBORDER or $80000000);
  clAppWorkSpace = XColor(COLOR_APPWORKSPACE or $80000000);
  clHighlight = XColor(COLOR_HIGHLIGHT or $80000000);
  clHighlightText = XColor(COLOR_HIGHLIGHTTEXT or $80000000);
  clBtnFace = XColor(COLOR_BTNFACE or $80000000);
  clBtnShadow = XColor(COLOR_BTNSHADOW or $80000000);
  clGrayText = XColor(COLOR_GRAYTEXT or $80000000);
  clBtnText = XColor(COLOR_BTNTEXT or $80000000);
  clInactiveCaptionText = XColor(COLOR_INACTIVECAPTIONTEXT or $80000000);
  clBtnHighlight = XColor(COLOR_BTNHIGHLIGHT or $80000000);
  cl3DDkShadow = XColor(COLOR_3DDKSHADOW or $80000000);
  cl3DLight = XColor(COLOR_3DLIGHT or $80000000);
  clInfoText = XColor(COLOR_INFOTEXT or $80000000);
  clInfoBk = XColor(COLOR_INFOBK or $80000000);

  clBlack = XColor($000000);
  clMaroon = XColor($000080);
  clGreen = XColor($008000);
  clOlive = XColor($008080);
  clNavy = XColor($800000);
  clPurple = XColor($800080);
  clTeal = XColor($808000);
  clGray = XColor($808080);
  clSilver = XColor($C0C0C0);
  clRed = XColor($0000FF);
  clLime = XColor($00FF00);
  clYellow = XColor($00FFFF);
  clBlue = XColor($FF0000);
  clFuchsia = XColor($FF00FF);
  clAqua = XColor($FFFF00);
  clLtGray = XColor($C0C0C0);
  clDkGray = XColor($808080);
  clWhite = XColor($FFFFFF);
  clNone = XColor($1FFFFFFF);
  clDefault = XColor($20000000);

type
   XGraphicObject = class
   private
      fHandle : Integer;
      fColor : XColor;
      fOnChange : XOnEvent;
      procedure Changed;
      procedure SetColor( Value : XColor );
      procedure ObjectCreating;
      procedure ObjectCreated;
   protected
      function GetHandle : Integer; virtual;
   public
      destructor Destroy; override;
      property Handle : Integer read GetHandle;
      property OnChange : XOnEvent read fOnChange write fOnChange;
      function ReleaseHandle : Integer;
      property Color : XColor read fColor write SetColor;
      procedure Assign( Value : XGraphicObject ); dynamic;
      procedure AssignHandle( NewHandle : Integer );
      function HandleAllocated : Boolean;
   end;

   XPenStyle = (psSolid, psDash, psDot, psDashDot, psDashDotDot, psClear,
    psInsideFrame);
   XPenMode = (pmBlack, pmWhite, pmNop, pmNot, pmCopy, pmNotCopy,
    pmMergePenNot, pmMaskPenNot, pmMergeNotPen, pmMaskNotPen, pmMerge,
    pmNotMerge, pmMask, pmNotMask, pmXor, pmNotXor);

   XPen = class( XGraphicObject )
   private
      fWidth: Integer;
      fStyle: XPenStyle;
      fMode : XPenMode;
      procedure SetWidth( Value : Integer );
      procedure SetStyle( Value : XPenStyle );
      procedure SetMode( Value : XPenMode );
      procedure Changed;
   protected
      function GetHandle : Integer; override;
   public
      constructor Create;
      property Handle;
      property Color;
      property Width : Integer read fWidth write SetWidth;
      property Style : XPenStyle read fStyle write SetStyle;
      property Mode : XPenMode read fMode write SetMode;
      property OnChange;
      procedure Assign( Value : XGraphicObject ); override;
   end;

   XBrushStyle = (bsSolid, bsClear, bsHorizontal, bsVertical,
    bsFDiagonal, bsBDiagonal, bsCross, bsDiagCross);

   XBrush = class( XGraphicObject )
   private
      fBitmap: HBitmap;
      fStyle: XBrushStyle;
      function GetHandle : Integer; override;
      procedure SetBitmap( Value : HBitmap );
      procedure SetStyle( Value : XBrushStyle );
   public
      constructor Create;
      destructor Destroy; override;
      property Handle;
      property Color;
      property Bitmap : HBitmap read fBitmap write SetBitmap;
      property Style : XBrushStyle read fStyle write SetStyle;
      property OnChange;
      procedure Assign( Value : XGraphicObject ); override;
   end;

   XFontStyles = (fsBold, fsItalic, fsUnderline, fsStrikeOut);
   XFontStyle = set of XFontStyles;
   XFontPitch = (fpDefault, fpVariable, fpFixed);
   XFontName = type string;
   XFontCharset = 0..255;

   XFont = class( XGraphicObject )
   private
      fHeight: Integer;
      fPitch: XFontPitch;
      fStyle: XFontStyle;
      fCharset: XFontCharset;
      fOrientation : Integer;
      fName: String;
      function GetHandle : Integer; override;
      procedure SetHeight( Value : Integer );
      procedure SetPitch( Value : XFontPitch );
      procedure SetStyle( Value : XFontStyle );
      procedure SetCharset( Value : XFontCharset );
      procedure SetOrientation( Value : Integer );
      procedure SetName( Value : String );
   public
      constructor Create;
      property Handle;
      property Color;
      property Height : Integer read fHeight write SetHeight;
      property Pitch : XFontPitch read fPitch write SetPitch;
      property Style : XFontStyle read fStyle write SetStyle;
      property Charset : XFontCharset read fCharset write SetCharset;
      property Orientation : Integer read fOrientation write SetOrientation;
      property Name : String read fName write SetName;
      procedure Assign( Value : XGraphicObject ); override;
   end;

var    // New TFont instances are intialized with the values in this structure:
  DefFontName : String = 'MS Sans Serif';
  DefFontHeight : Integer = 0;
  DefFontPitch : XFontPitch = fpDefault;
  DefFontStyle : XFontStyle = [];
  DefFontCharset : XFontCharset = DEFAULT_CHARSET;
  DefFontOrientation : Integer = 0;
  DefFontColor : XColor = clWindowText;

type
   XGraphicsManager = class
   private
      fOnObjectCreating : XOnEvent;
      fOnObjectCreated : XOnEvent;
   public
      property OnObjectCreating : XOnEvent read fOnObjectCreating write fOnObjectCreating;
      property OnObjectCreated : XOnEvent read fOnObjectCreated write fOnObjectCreated;
   end;

var GlobalGraphicsManager : XGraphicsManager;

function Color2RGB( Color : XColor ) : XColor;
function ColorsMix( Color1, Color2 : XColor ) : XColor;

implementation

function Color2RGB( Color : XColor ) : XColor;
begin
  if Color < 0 then
    Result := GetSysColor(Color and $FF) else
    Result := Color;
end;

function ColorsMix( Color1, Color2 : XColor ) : XColor; assembler;
asm
   CALL Color2Rgb
   XCHG EAX, EDX
   CALL Color2Rgb
   AND EAX, 0FEFEFEh
   AND EDX, 0FEFEFEh
   SHR EAX, 1
   SHR EDX, 1
   ADD EAX, EDX
end;


///////////////////////////////////////////////////////////////
//                                                           //
//   GGGGGG                            HH                    //
//  GG    GG                           HH        II          //
//  GG        RR RR    AAAA    PPPPP   HHHHHH         CCCCC  //
//  GG    GGG RRR  R  A    A   PP  PP  HH   HH   II  CC   CC //
//  GG    GG  RR         AAA   PPPPP   HH   HH   II  CC      //
//  GG    GG  RR        A   A  PP      HH   HH   II  CC   CC //
//   GGGGGG   RR        AAA A  PP      HH   HH   II   CCCCC  //
//                                                           //
///////////////////////////////////////////////////////////////

destructor XGraphicObject.Destroy;
begin
   if fHandle <> 0 then
      DeleteObject( fHandle );
   inherited;
end;

procedure XGraphicObject.Changed;
begin
   if Assigned( fOnChange ) then
      fOnChange( Self );
   if fHandle <> 0 then
      DeleteObject( fHandle );
   fHandle := 0;
end;

function XGraphicObject.GetHandle : Integer;
begin
   Result := fHandle;
end;

function XGraphicObject.ReleaseHandle : Integer;
begin
   Changed;
   Result := fHandle;
   fHandle := 0;
end;

procedure XGraphicObject.SetColor( Value : XColor );
begin
   if Color2RGB( fColor ) = Color2RGB( Value ) then Exit;
   fColor := Value;
   Changed;
end;

procedure XGraphicObject.ObjectCreating;
begin
   if GlobalGraphicsManager <> nil then
   if Assigned( GlobalGraphicsManager.fOnObjectCreating ) then
      GlobalGraphicsManager.fOnObjectCreating( Self );
end;

procedure XGraphicObject.ObjectCreated;
begin
   if GlobalGraphicsManager <> nil then
   if Assigned( GlobalGraphicsManager.fOnObjectCreated ) then
      GlobalGraphicsManager.fOnObjectCreated( Self );
end;

procedure XGraphicObject.Assign( Value : XGraphicObject );
begin
   Changed;
   if fHandle <> 0 then
      DeleteObject( fHandle );
   fHandle := 0;
   fColor := Value.fColor;
end;

procedure XGraphicObject.AssignHandle( NewHandle : Integer );
begin
   if fHandle <> 0 then
      DeleteObject( fHandle );
   fHandle := NewHandle;
end;

function XGraphicObject.HandleAllocated : Boolean;
begin
   Result := fHandle <> 0;
end;

///////////////////////////////////////////////////////////////////////////
//                                                                       //
//  ////////////////\     ///////////////////    ///////\        //////  //
//  //////////////////    ///////////////////    ////////\       //////  //
//  //////      //////    //////                 /////////\      //////  //
//  //////      //////    //////                 //////  \\\     //////  //
//  //////     //////     //////                 //////   \\\    //////  //
//  ///////////////       //////////////         //////    \\\   //////  //
//  /////////////         //////////////         //////     \\\  //////  //
//  //////                //////                 //////      \\\ //////  //
//  //////                //////                 //////       \\\//////  //
//  //////                ///////////////////    //////        \\\/////  //
//  //////                ///////////////////    //////         \\/////  //
//                                                                       //
///////////////////////////////////////////////////////////////////////////

constructor XPen.Create;
begin
   fMode := pmCopy;
end;

function XPen.GetHandle : Integer;
const
  PenStyles: array[ XPenStyle ] of Word =
    (PS_SOLID, PS_DASH, PS_DOT, PS_DASHDOT, PS_DASHDOTDOT, PS_NULL,
     PS_INSIDEFRAME);
var
  LogPen: TLogPen;
begin
   if fHandle = 0 then
   begin
      ObjectCreating;
      if fHandle = 0 then
      with LogPen do
      begin
         lopnStyle := PenStyles[ Style ];
         lopnWidth.X := Width;
         lopnColor := Color2RGB( Color );
         fHandle := CreatePenIndirect( LogPen );
      end;
      ObjectCreated;
   end;
   Result := fHandle;
end;

procedure XPen.Changed;
begin
   if Assigned( fOnChange ) then
      fOnChange( Self );
   if fHandle <> 0 then
   begin
      DeleteObject( fHandle );
      fHandle := 0;
   end;
end;

procedure XPen.SetWidth( Value : Integer );
begin
   if fWidth = Value then Exit;
   fWidth := Value;
   Changed;
end;

procedure XPen.SetStyle( Value : XPenStyle );
begin
   if fStyle = Value then Exit;
   fStyle := Value;
   Changed;
end;

procedure XPen.SetMode( Value : XPenMode );
begin
   if fMode = Value then Exit;
   fMode := Value;
   Changed;
end;

procedure XPen.Assign( Value : XGraphicObject );
var Pen : XPen;
begin
   if not (Value is XPen) then Exit;
   Pen := Value as XPen;
   if fHandle <> 0 then
      if Pen.fHandle = fHandle then Exit;
   if fWidth = Pen.Width then
   if fStyle = Pen.Style then
   if fMode = Pen.Mode then Exit;
   inherited Assign( Value );
   fWidth := Pen.fWidth;
   fStyle := Pen.fStyle;
   fMode := Pen.fMode;
end;

///////////////////////////////////////////////////////////////////////////////////
//                                                                               //
//  //////////\    //////////\      /////   /////     //////\     /////   /////  //
//  ///////////\   ///////////\     /////   /////   //////////\   /////   /////  //
//  /////   ////|  /////   ////|    /////   /////  /////   ////|  /////   /////  //
//  /////   /////  /////   /////    /////   /////  /////          /////   /////  //
//  /////   /////  /////   /////    /////   /////  /////          /////   /////  //
//  ///////////    ///////////      /////   /////  ///////////    /////////////  //
//  ///////////\   ////////\        /////   /////   //////////\   /////////////  //
//  /////   ////\  ///// ////\      /////   /////          ////\  /////   /////  //
//  /////   /////  /////  ////\     /////   /////          /////  /////   /////  //
//  /////   /////  /////   ////\    /////   /////          /////  /////   /////  //
//  /////   ////   /////    ////\   /////   /////  /////   ////   /////   /////  //
//  ///////////    /////     ////\   ///////////    //////////    /////   /////  //
//  //////////     /////      ////\    ////////      ////////     /////   /////  //
//                                                                               //
///////////////////////////////////////////////////////////////////////////////////                                                                                 //

constructor XBrush.Create;
begin
   inherited;
   fColor := clBtnFace;
   fStyle := bsSolid;
end;

destructor XBrush.Destroy;
begin
   if fBitmap <> 0 then
      DeleteObject( fBitmap );
   inherited;
end;

function XBrush.GetHandle : Integer;
var
  LogBrush: TLogBrush;
begin
    if fHandle = 0 then
    begin
       ObjectCreating;
       if fHandle = 0 then
       begin
          with LogBrush do
          begin
            if fBitmap <> 0 then
            begin
              lbStyle := BS_PATTERN;
              lbHatch := fBitmap;
            end
               else
            begin
              lbHatch := 0;
              case fStyle of
                bsSolid: lbStyle := BS_SOLID;
                bsClear: lbStyle := BS_HOLLOW;
              else
                lbStyle := BS_HATCHED;
                lbHatch := Ord( fStyle ) - Ord( bsHorizontal );
              end;
            end;
            lbColor := Color2RGB( fColor );
          end;
          fHandle := CreateBrushIndirect(LogBrush);
       end;
       ObjectCreated;
    end;
    Result := fHandle;
end;

procedure XBrush.SetBitmap( Value : HBitmap );
begin
   if fBitmap = Value then Exit;
   fBitmap := Value;
   Changed;
end;

procedure XBrush.SetStyle( Value : XBrushStyle );
begin
   if fStyle = Value then Exit;
   fStyle := Value;
   Changed;
end;

procedure XBrush.Assign( Value : XGraphicObject );
var Brush : XBrush;
begin
   if not (Value is XBrush) then Exit;
   Brush := Value as XBrush;
   if Brush.fHandle = fHandle then
   if fHandle <> 0 then Exit;
   if fBitmap = Brush.Bitmap then
   if fStyle = Brush.Style then Exit;
   inherited Assign( Value );
   fBitmap := Brush.fBitmap;
   fStyle := Brush.fStyle;
end;

///////////////////////////////////////////////////////////////////////////
//                                                                       //
//  /////////////    /////////      /////\        /////   /////////////  //
//  /////////////   ///////////     /////\        /////   /////////////  //
//  /////          /////   /////    //////\       /////       /////      //
//  /////          /////   /////    ///////\      /////       /////      //
//  /////          /////   /////    ///// \\\     /////       /////      //
//  ///////////    /////   /////    /////  \\\    /////       /////      //
//  ///////////    /////   /////    /////   \\\   /////       /////      //
//  /////          /////   /////    /////    \\\  /////       /////      //
//  /////          /////   /////    /////     \\\ /////       /////      //
//  /////          /////   /////    /////      \\\\////       /////      //
//  /////          /////   /////    /////       \\\////       /////      //
//  /////           ///////////     /////        \/////       /////      //
//  /////             ////////      /////         /////       /////      //
//                                                                       //
///////////////////////////////////////////////////////////////////////////

constructor XFont.Create;
begin
   inherited;
   fName := DefFontName;
   fCharSet := DefFontCharset;
   fPitch := DefFontPitch;
   fStyle := DefFontStyle;
   fHeight := DefFontHeight;
   fColor := DefFontColor;
   fOrientation := DefFontOrientation;
end;

function XFont.GetHandle : Integer;
var
  LogFont: TLogFont;
begin
   if fHandle = 0 then
   begin
      ObjectCreating;
      if fHandle = 0 then
      with LogFont do
      begin
         lfHeight := Height;
         lfWidth := 0; { have font mapper choose }
         lfEscapement := Orientation; { not only straight fonts }
         lfOrientation := Orientation; { can be rotated }
         if fsBold in Style then
            lfWeight := FW_BOLD
         else
            lfWeight := FW_NORMAL;
         lfItalic := Byte( fsItalic in Style );
         lfUnderline := Byte( fsUnderline in Style );
         lfStrikeOut := Byte( fsStrikeOut in Style );
         lfCharSet := Byte( Charset );
         if Name = 'Default' then  // do not localize
            StrCopy( lfFaceName, PChar( DefFontName ) )
         else
            StrCopy( lfFaceName, PChar( Name ) );
         lfQuality := DEFAULT_QUALITY;
         { Everything else as default }
         lfOutPrecision := OUT_DEFAULT_PRECIS;
         lfClipPrecision := CLIP_DEFAULT_PRECIS;
         case Pitch of
            fpVariable: lfPitchAndFamily := VARIABLE_PITCH;
            fpFixed: lfPitchAndFamily := FIXED_PITCH;
         else
            lfPitchAndFamily := DEFAULT_PITCH;
         end;
         fHandle := CreateFontIndirect(LogFont);
      end;
      ObjectCreated;
   end;
   Result := fHandle;
end;

procedure XFont.SetHeight( Value : Integer );
begin
   if fHeight = Value then Exit;
   fHeight := Value;
   Changed;
end;

procedure XFont.SetPitch( Value : XFontPitch );
begin
   if fPitch = Value then Exit;
   fPitch := Value;
   Changed;
end;

procedure XFont.SetStyle( Value : XFontStyle );
begin
   if fStyle = Value then Exit;
   fStyle := Value;
   Changed;
end;

procedure XFont.SetCharset( Value : XFontCharset );
begin
   if fCharset = Value then Exit;
   fCharset := Value;
   Changed;
end;

procedure XFont.SetOrientation( Value : Integer );
begin
   if fOrientation = Value then Exit;
   fOrientation := Value;
   Changed;
end;

procedure XFont.SetName( Value : String );
begin
   if fName = Value then Exit;
   fName := Value;
   Changed;
end;

procedure XFont.Assign( Value : XGraphicObject );
var Font : XFont;
begin
   if not (Value is XFont) then Exit;
   Font := Value as XFont;
   if Font.fHandle = fHandle then
   if fHandle <> 0 then Exit;
   if fHeight = Font.fHeight then
   if fPitch = Font.fPitch then
   if fStyle = Font.fStyle then
   if fOrientation = Font.fOrientation then
   if fCharset = Font.fCharset then
   if fName = Font.fName then Exit;
   inherited Assign( Value );
   fHeight := Font.fHeight;
   fPitch := Font.fPitch;
   fStyle := Font.fStyle;
   fOrientation := Font.fOrientation;
   fCharset := Font.fCharset;
   fName := Font.fName;
   Changed;
end;

end.
