unit XMFCControls;

interface

uses Windows, Messages, XClasses, XVisuals, XWindows, XForms, XGraphics,
     XStrUtils, XApplets;

type
   XMfcControl = class;

   XMfcWindow = class( XWindow )
   private
     FMasterControl: XMfcControl;
   protected
     procedure CreateParams( var Params : TCreateParams ); override;
     procedure SubclassWindow( var Params : TCreateParams; ClassName : PChar );
     function GetParentWnd( NeedHandle : Boolean ) : HWnd; override;
   public
     constructor Create( AParent : XClass ); override;
     function WndProc( var Msg : TMsg ) : Integer; override;
     property MasterControl : XMfcControl read FMasterControl;
   end;

   XMfcControl = class( XControl )
   private
     fWindow : XMfcWindow;
    FFont: XFont;
    FBrush: XBrush;
    procedure SetFont(const Value: XFont);
    procedure SetBrush(const Value: XBrush);
   protected
     property Font : XFont read FFont write SetFont;
     property Brush : XBrush read FBrush write SetBrush;
     procedure SetBoundsRect( Value : TRect ); override;
      function GetBrushColor : XColor;
      procedure SetBrushColor( Value : XColor );
      procedure CreateParams( var Params : TCreateParams ); virtual;
      procedure FontChanged( Sender : TObject ); dynamic;
      procedure BrushChanged( Sender : TObject ); dynamic;
   public
     constructor Create( AParent : XClass ); override;
     destructor Destroy; override;
     function Windowed : Boolean; override;
     function GetWindowHandle : HWnd; override;
     function CreateWindow : Boolean; override;
     function WndProc( var Msg : TMsg ) : Integer; override;
     property Enabled;
     property Visible;
     property Showing;
     property Color : XColor read GetBrushColor write SetBrushColor;
     procedure DoPaint( DC : HDC; Rect : TRect ); override;
     procedure Invalidate; override;
     function ClientRect : TRect; override;
   end;

   XMfcControlsManager = class
   private
    FOn_MfcControlMessage: XOnFormMessage;
   protected
   public
     property On_MfcControlMessage : XOnFormMessage read FOn_MfcControlMessage write FOn_MfcControlMessage;
   end;

var GlobalMfcControlsManager : XMfcControlsManager;

type
   XWCustomStatic = class( XMfcControl )
   private
    function GetCaption: String;
    procedure SetCaption(const Value: String); virtual;
   protected
     procedure CreateParams( var Params : TCreateParams ); override;
     property Brush;
     property Font;
     property Caption : String read GetCaption write SetCaption;
   public
     constructor Create( AParent : XClass ); override;
     function CanFocus : Boolean; override;
   end;

   XWLabel = class( XWCustomStatic )
   private
    FAutoSize: Boolean;
    procedure SetAutoSize(const Value: Boolean);
    procedure SetCaption( const Value : String ); override;
    procedure AdjustSize;
   protected
     procedure FontChanged( Sender : TObject ); override;
   public
     constructor Create( AParent : XClass ); override;
     property Font;
     property Brush;
     property Color;
     property Caption;
     property AutoSize : Boolean read FAutoSize write SetAutoSize;
   end;
   CLabel = XWLabel;

   XWCustomPaint = class( XMfcControl )
   private
    FOnPaint: XOnEvent;
    FOnEraseBkgnd: XOnEvent;
   protected
     property Brush;
     property Color;
     property OnPaint : XOnEvent read FOnPaint write FOnPaint;
     property OnEraseBkgnd : XOnEvent read FOnEraseBkgnd write FOnEraseBkgnd;
   public
     function WndProc( var Msg : TMsg ) : Integer; override;
     function CanFocus : Boolean; override;
   end;
   CCustomPaint = XWCustomPaint;

   XWPaint = class( XWCustomPaint )
   private
   protected
   public
     constructor Create( AParent : XClass ); override;
     property Brush;
     property Color;
     property OnPaint;
     property OnEraseBkgnd;
   end;
   CPaint = XWPaint;

implementation

{ XMfcControl }

procedure XMfcControl.BrushChanged(Sender: TObject);
begin
  // nothing here
end;

function XMfcControl.ClientRect: TRect;
begin
  Result := FWindow.ClientRect;
end;

constructor XMfcControl.Create(AParent: XClass);
begin
  inherited;
  FFont := XFont.Create;
  FBrush := XBrush.Create;
  FBrush.Color := Color;
  FFont.OnChange := FontChanged;
  FBrush.OnChange := BrushChanged;
  fWindow := XMfcWindow.Create( Self );
end;

procedure XMfcControl.CreateParams(var Params: TCreateParams);
begin
  // nothing here - to allow inherited call only
end;

function XMfcControl.CreateWindow: Boolean;
begin
  Result := fWindow.CreateWindow;
  SetFont( FFont );
  SetBrush( FBrush );
end;

destructor XMfcControl.Destroy;
begin
  fWindow.Free;
  inherited;
end;

procedure XMfcControl.DoPaint(DC: HDC; Rect: TRect);
begin
  PaintChilds( DC, Rect );
  if FWindow.Handle <> 0 then
  begin
     InvalidateRect( FWindow.Handle, nil, True );
     UpdateWindow( FWindow.Handle );
  end;
  ExcludeUpdRect( DC, Rect );
end;

procedure XMfcControl.FontChanged(Sender: TObject);
begin
  // nothing here
end;

function XMfcControl.GetBrushColor: XColor;
begin
   Result := FBrush.Color;
end;

function XMfcControl.GetWindowHandle: HWnd;
begin
   Result := fWindow.Handle;
end;

procedure XMfcControl.Invalidate;
begin
   inherited;
   if FWindow.Handle <> 0 then
      InvalidateRect( FWindow.Handle, nil, True );
end;

procedure XMfcControl.SetBoundsRect(Value: TRect);
begin
  inherited;
  FWindow.BoundsRect := ControlRect;
end;

procedure XMfcControl.SetBrush(const Value: XBrush);
begin
  FBrush.Assign( Value );
  Invalidate;
end;

procedure XMfcControl.SetBrushColor(Value: XColor);
begin
   if FBrush.Color <> Value then
   begin
      FBrush.Color := Value;
      Invalidate;
   end;
end;

procedure XMfcControl.SetFont(const Value: XFont);
begin
  FFont.Assign( Value );
  if FWindow.Handle <> 0 then
     SendMessage( FWindow.Handle, WM_SETFONT, FFont.Handle, 1 );
  Invalidate;
end;

function XMfcControl.Windowed: Boolean;
begin
   Result := True;
end;

function XMfcControl.WndProc(var Msg: TMsg): Integer;
var Handled : Boolean;
begin
  if Assigned( GlobalMfcControlsManager.On_MfcControlMessage ) then
  begin
     Handled := False;
     GlobalMfcControlsManager.On_MfcControlMessage( Self, Msg, Result, Handled );
     if Handled then Exit;
  end;
  case Msg.message of
   WM_CTLCOLORMSGBOX..WM_CTLCOLORSTATIC:
       Result := SendMessage(Msg.LParam, CN_BASE + Msg.message, Msg.WParam, Msg.LParam);
   CN_CTLCOLORMSGBOX..CN_CTLCOLORSTATIC:
       begin
         SetTextColor(Msg.WParam, Color2RGB(FFont.Color));
         SetBkColor(Msg.WParam, Color2RGB(FBrush.Color));
         Result := FBrush.Handle;
       end;
   else Result := FWindow.CallDefWndProc( Msg );
  end;
end;

{ XMfcWindow }

constructor XMfcWindow.Create(AParent: XClass);
begin
  inherited Create( nil );
  fMasterControl := AParent as XMfcControl;
  Style := WS_VISIBLE or WS_CLIPCHILDREN or WS_CLIPSIBLINGS or WS_CHILD;
  if fMasterControl.CanFocus then
     Style := Style or WS_TABSTOP;
end;

procedure XMfcWindow.CreateParams(var Params: TCreateParams);
begin
  inherited;
  fMasterControl.CreateParams( Params );
end;

function XMfcWindow.GetParentWnd(NeedHandle: Boolean): HWnd;
begin
   Result := FMasterControl.GetParentWnd( NeedHandle );
end;

procedure XMfcWindow.SubclassWindow(var Params : TCreateParams; ClassName: PChar );
begin
  CreateSubclass( Params, ClassName );
end;

function XMfcWindow.WndProc(var Msg: TMsg): Integer;
begin
  Result := fMasterControl.WndProc( Msg );
end;

{ XWCustomStatic }

function XWCustomStatic.CanFocus: Boolean;
begin
  Result := False;
end;

constructor XWCustomStatic.Create(AParent: XClass);
begin
  inherited;
  Height := DefFontHeight;
  Width := 64;
end;

procedure XWCustomStatic.CreateParams(var Params: TCreateParams);
begin
  inherited;
  StrCopy( Params.WinClassName, PChar( String( ClassName ) ) );
  FWindow.SubclassWindow( Params, 'STATIC' );
end;

function XWCustomStatic.GetCaption: String;
begin
  Result := FWindow.Caption;
end;

procedure XWCustomStatic.SetCaption(const Value: String);
begin
  if Caption = Value then Exit;
  FWindow.Invalidate;
  FWindow.Caption := Value;
  Invalidate;
end;

{ XWLabel }

procedure XWLabel.AdjustSize;
var R : TRect;
    DC : HDC;
    OldFont : HFont;
    Sz : TSize;
begin
   R := BoundsRect;
   DC := CreateCompatibleDC( 0 );
   OldFont := SelectObject( DC, Font.Handle );
     Windows.GetTextExtentPoint( DC, PChar( Caption ), Length( Caption ), Sz );
   SelectObject( DC, OldFont );
   DeleteDC( DC );
   R.Right := R.Left + Sz.cx;
   R.Bottom := R.Top + Sz.cy;
   BoundsRect := R;
end;

constructor XWLabel.Create(AParent: XClass);
begin
   inherited;
   FAutoSize := True;
end;

procedure XWLabel.FontChanged(Sender: TObject);
begin
  if FAutoSize then
     AdjustSize;
end;

procedure XWLabel.SetAutoSize(const Value: Boolean);
begin
  if FAutoSize = Value then Exit;
  FAutoSize := Value;
  if FAutoSize then
     AdjustSize;
end;

procedure XWLabel.SetCaption(const Value: String);
begin
   inherited;
   if FAutoSize then
      AdjustSize;
end;

{ XWCustomPaint }

{ XWCustomPaint }


{ XWCustomPaint }

function XWCustomPaint.CanFocus: Boolean;
begin
  Result := False;
end;

function XWCustomPaint.WndProc(var Msg: TMsg): Integer;
begin
  case Msg.message of
  WM_PAINT:
    begin
       if Assigned( FOnPaint ) then
          FOnPaint( Self );
       Result := 0;
    end;
  WM_ERASEBKGND:
    begin
       if Assigned( FOnEraseBkgnd ) then
          FOnEraseBkgnd( Self )
       else
          inherited WndProc( Msg );
       Result := 1;
    end;
  else
    Result := inherited WndProc( Msg );
  end;
end;

{ XWPaint }

constructor XWPaint.Create(AParent: XClass);
var R : TRect;
begin
   inherited;
   R := BoundsRect;
   R.Right := R.Left + 64;
   R.Bottom := R.Top + 64;
   BoundsRect := R;
end;

initialization

   GlobalMfcControlsManager := XMfcControlsManager.Create;

finalization

   GlobalMfcControlsManager.Free;

end.
