unit XTimers;

interface

uses Windows, Messages, XWindows, XClasses, XApplets;

type
   ZTimer = class
   private
    FHandle : Integer;
    FEnabled: Boolean;
    FInterval: Integer;
    FOnTimer: XOnEvent;
    procedure SetEnabled(const Value: Boolean);
    procedure SetInterval(const Value: Integer);
    procedure SetOnTimer(const Value: XOnEvent);
    procedure Timer;
   protected
   public
      constructor Create;
      destructor Destroy; override;
      property Handle : Integer read FHandle;
      property Enabled : Boolean read FEnabled write SetEnabled;
      property Interval : Integer read FInterval write SetInterval;
      property OnTimer : XOnEvent read FOnTimer write SetOnTimer;
   end;

   XTimer = class( XChild )
   private
     fTimer : ZTimer;
    function GetEnabled: Boolean;
    function GetHandle: Integer;
    function GetInterval: Integer;
    function GetOnTimer: XOnEvent;
    procedure SetEnabled(const Value: Boolean);
    procedure SetInterval(const Value: Integer);
    procedure SetOnTimer(const Value: XOnEvent);
   protected
   public
     constructor Create( AParent : XClass ); override;
     destructor Destroy; override;
     property Handle : Integer read GetHandle;
     property Enabled : Boolean read GetEnabled write SetEnabled;
     property Interval : Integer read GetInterval write SetInterval;
     property OnTimer : XOnEvent read GetOnTimer write SetOnTimer;
   end;

implementation

{ XTimer }

procedure TimerProc( Wnd : HWnd; Msg : Integer; Id : Integer; CurrentTime : DWord );
          stdcall;
var T : ZTimer;
begin
   T := Pointer( Id );
   T.Timer;
end;

constructor ZTimer.Create;
begin
  inherited;
  FInterval := 1000;
end;

destructor ZTimer.Destroy;
begin
  Enabled := False;
  inherited;
end;

procedure ZTimer.SetEnabled(const Value: Boolean);
begin
  if FEnabled = Value then Exit;
  if Value then
  begin
     if Applet.Handle = 0 then
        if not Applet.CreateAppButton then Exit;
     FHandle := SetTimer( Applet.Handle, Integer( Pointer( Self ) ),
                          fInterval, @TimerProc );
  end
     else
  begin
     if FHandle <> 0 then
     begin
        KillTimer( Applet.Handle, FHandle );
        FHandle := 0;
     end;
  end;
  FEnabled := Value;
end;

procedure ZTimer.SetInterval(const Value: Integer);
var WasEnabled : Boolean;
begin
  if FInterval = Value then Exit;
  WasEnabled := Enabled;
  Enabled := False;
  FInterval := Value;
  Enabled := WasEnabled;
end;

procedure ZTimer.SetOnTimer(const Value: XOnEvent);
begin
  FOnTimer := Value;
end;

procedure ZTimer.Timer;
begin
   if Assigned( fOnTimer ) then
      fOnTimer( Self );
end;

function XTimer.GetEnabled: Boolean;
begin
   Result := fTimer.Enabled;
end;

function XTimer.GetHandle: Integer;
begin
   Result := fTimer.Handle;
end;

function XTimer.GetInterval: Integer;
begin
   Result := fTimer.Interval;
end;

function XTimer.GetOnTimer: XOnEvent;
begin
   Result := fTimer.FOnTimer;
end;

procedure XTimer.SetEnabled(const Value: Boolean);
begin
   fTimer.SetEnabled( Value );
end;

procedure XTimer.SetInterval(const Value: Integer);
begin
   fTimer.SetInterval( Value );
end;

procedure XTimer.SetOnTimer(const Value: XOnEvent);
begin
   fTimer.fOnTimer := Value;
end;

{ XTimer }

constructor XTimer.Create(AParent: XClass);
begin
   inherited;
   fTimer := ZTimer.Create;
end;

destructor XTimer.Destroy;
begin
   fTimer.Free;
   inherited;
end;

end.
