unit XWindows;

interface

uses Windows, Messages, XClasses, XStrUtils, XVisuals, XGraphics;

const
   CM_BASE = $B000;
   CM_ACTIVATE   = CM_BASE + 0;
   CM_DEACTIVATE = CM_BASE + 1;
   CM_ENTER      = CM_BASE + 2;
   CM_RELEASE    = CM_BASE + 3;

type
   XOnMessage = procedure( var Msg : TMsg; var Rslt : Integer; var Handled : Boolean )
                           of object;

   XColor = Integer;

   XOnEventAccept = procedure( Sender : TObject; var Accept : Boolean ) of object;
   XWindowState = ( xwsNormal, xwsMinimized, xwsMaximized );

  TCreateParams = record
    Caption: PChar;
    Style: Longint;
    ExStyle: Longint;
    X, Y: Integer;
    Width, Height: Integer;
    WndParent: HWnd;
    Param: Pointer;
    WindowClass: TWndClass;
    WinClassName: array[0..63] of Char;
  end;

   XWindow = class( XVisual )
   private
      fHandle : HWnd;
      fDestroying : Boolean;
      fClsStyle : Integer;
      fStyle : Integer;
      fExStyle : Integer;
      fCursor : HCursor;
      fIcon : HIcon;
      fMenu : HMenu;
      fCaption : PChar;
      fOnClose : XOnEventAccept;
      fWindowState : XWindowState;
      fShowAction : Integer;
      fCanvas : Pointer;
      fDefWndProc : Pointer;
      function GetParentWindow : HWnd;
      procedure SetEnabled( Value : Boolean ); override;
      function GetVisible : Boolean; override;
      procedure SetVisible( Value : Boolean ); override;
      procedure SetClsStyle( Value : Integer );
      procedure SetStyle( Value : Integer );
      procedure SetExStyle( Value : Integer );
      procedure SetCursor( Value : HCursor );
      procedure SetIcon( Value : HIcon );
      procedure SetMenu( Value : HMenu );
    function GetCaption: String;
      procedure SetCaption( Value : String );
      procedure SetColor( Value : XColor ); override;
      function GetBoundsRect : TRect; override;
      procedure SetBoundsRect( Value : TRect ); override;
      procedure SetWindowState( Value : XWindowState );
      function GetHandle : HWnd;
   protected
      function GetIcon : HIcon; virtual;
      function GetWindowHandle : HWnd; override;
      procedure CreateParams( var Params : TCreateParams ); dynamic;
      procedure CreateSubclass( var Params : TCreateParams; ControlClassName : PChar );
   public
      constructor Create( AParent : XClass ); override;
      destructor Destroy; override;
      function CreateWindow : Boolean; override;
      property Handle : HWnd read GetHandle;
      property Enabled;
      property ParentWindow : HWnd read GetParentWindow;
      property ClsStyle : Integer read fClsStyle write SetClsStyle;
      property Style : Integer read fStyle write SetStyle;
      property ExStyle : Integer read fExStyle write SetExStyle;
      property Cursor : HCursor read fCursor write SetCursor;
      property Icon : HIcon read GetIcon write SetIcon;
      property Menu : HMenu read fMenu write SetMenu;
      property Caption : String read GetCaption write SetCaption;
      property Visible;
      procedure Show;
      procedure Hide;
      property Color;
      property BoundsRect;
      property Left;
      property Top;
      property Width;
      property Height;
      property OnClose : XOnEventAccept read fOnClose write fOnClose;
      property WindowState : XWindowState read fWindowState write SetWindowState;
      function ClientToScreen( P : TPoint ) : TPoint;
      function ScreenToClient( P : TPoint ) : TPoint;
      function ClientRect : TRect; override;
      property Canvas : Pointer read fCanvas write fCanvas;
      procedure Invalidate; override;
      function Windowed : Boolean; override;
      function WndProc( var Msg : TMsg ) : Integer; override;
      function CallDefWndProc( var Msg : TMsg ) : Integer;
   end;

   XOnPaint = procedure( Sender : TObject; DC : HDC ) of object;

   XWindowsManager = class
   private
      fOnCreateWindow : XOnEvent;
      fOnDestroyWindow : XOnEvent;
      fOnBeginPaint : XOnPaint;
      fOnEndPaint : XOnPaint;
   public
      property OnCreateWindow : XOnEvent read fOnCreateWindow write fOnCreateWindow;
      property OnDestroyWindow : XOnEvent read fOnDestroyWindow write fOnDestroyWindow;
      property OnBeginPaint : XOnPaint read fOnBeginPaint write fOnBeginPaint;
      property OnEndPaint : XOnPaint read fOnEndPaint write fOnEndPaint;
   end;

var GlobalWindowsManager : XWindowsManager;

function WndFunc( W : HWnd; Msg : Cardinal; wParam, lParam : Integer )
                                    : Integer; stdcall;

implementation

uses XApplets;

{
procedure ChangeWindowFlags( W : HWnd; FlagIndex : Integer; ExcludeFlags, IncludeFlags : Integer );
var Flags : Integer;
begin
   if W = 0 then Exit;
   Flags := (GetWindowLong( W, FlagIndex ) and not ExcludeFlags) or IncludeFlags;
   SetWindowLong( W, FlagIndex, Flags );
end;
}

var CreatingWindow : XWindow;

function WndFunc( W : HWnd; Msg : Cardinal; wParam, lParam : Integer )
                                    : Integer; stdcall;
var M : TMsg;
begin
   if CreatingWindow <> nil then
   if W <> 0 then
   begin
      CreatingWindow.fHandle := W;
      CreatingWindow := nil;
   end;
   M.hwnd := W;
   M.message := Msg;
   M.wParam := wParam;
   M.lParam := lParam;
   Result := Applet.WndProc( M );
end;

{ XWindow implementation }

constructor XWindow.Create( AParent : XClass );
begin
   inherited Create( AParent );
   if Applet <> nil then
      Applet.AddWindowed( Self );
   fClsStyle := {CS_DBLCLKS or} CS_OWNDC {or CS_HREDRAW or CS_VREDRAW};
   fStyle := WS_VISIBLE or WS_CLIPCHILDREN or WS_CLIPSIBLINGS or
             WS_CAPTION or WS_SYSMENU or WS_MINIMIZEBOX or WS_MAXIMIZEBOX or
             WS_BORDER or WS_THICKFRAME;
   fExStyle := WS_EX_CONTROLPARENT;
   inherited SetVisible( True );
   inherited SetEnabled( True );
   Color := clBtnFace;
   fCursor := LoadCursor( 0, IDC_ARROW );
   if Assigned( GlobalWindowsManager.fOnCreateWindow ) then
      GlobalWindowsManager.fOnCreateWindow( Self );
end;

destructor XWindow.Destroy;
var I : Integer;
    W : XClass;
begin
   if Assigned( GlobalWindowsManager.fOnDestroyWindow ) then
      GlobalWindowsManager.fOnDestroyWindow( Self );

   for I := ChildsCount - 1 downto 0 do
   begin
      W := Childs[ I ];
      W.Free;
   end;
   if not fDestroying then
   if Handle <> 0 then
   begin
      I := Handle;
      //fHandle := 0;
      SendMessage( I, WM_SETICON, 1, 0 );
      if not DestroyWindow( I ) then
         Exit;
      fHandle := 0;
   end;
   Applet.DeleteWindowed( Self );
   if FCaption <> nil then
      FreeMem( FCaption );
   inherited;
end;

procedure XWindow.SetEnabled( Value : Boolean );
begin
   if Enabled = Value then Exit;
   inherited SetEnabled( Value );
   if Enabled then
      fStyle := fStyle and not WS_DISABLED
   else
      fStyle := fStyle or WS_DISABLED;
   if fHandle <> 0 then
      EnableWindow( fHandle, Enabled );
end;

function XWindow.GetParentWindow : HWnd;
begin
   Result := 0;
   if Parent = nil then Exit;
   if Parent is XWindow then
   begin
      if (Parent as XWindow).Handle = 0 then
         (Parent as XWindow).CreateWindow;
      Result := (Parent as XWindow).Handle;
   end;
end;

function XWindow.CreateWindow : Boolean;
var TempClass : TWndClass;
    Params : TCreateParams;
    ClassRegistered : Boolean;
begin
   Result := False;
   if Applet.Handle = 0 then
   if Applet <> Self then
   if not Applet.CreateWindow then Exit;
   CreateParams( Params );
   if fHandle <> 0 then Exit;
   FDefWndProc := Params.WindowClass.lpfnWndProc;
   if (Params.WndParent = 0) and (Params.Style and WS_CHILD <> 0) then Exit;
   ClassRegistered := GetClassInfo( Params.WindowClass.hInstance,
                      Params.WinClassName, TempClass );
   if not ClassRegistered then
   begin
      Params.WindowClass.lpszClassName := Params.WinClassName;
      Params.WindowClass.lpfnWndProc := @WndFunc;
      if RegisterClass( Params.WindowClass ) = 0 then Exit;
   end;
   CreatingWindow := Self;
   fHandle := CreateWindowEx( Params.ExStyle, Params.WinClassName,
                              Params.Caption, Params.Style, Params.X, Params.Y,
                              Params.Width, Params.Height, Params.WndParent,
                              fMenu, Params.WindowClass.hInstance, Params.Param );
   if fHandle = 0 then Exit;
   SendMessage( fHandle, WM_SETICON, ICON_BIG, GetIcon );
   CreateChildWindows;
   Result := True;
end;

function XWindow.Windowed : Boolean;
begin
   Result := True;
end;

function XWindow.WndProc( var Msg : TMsg ) : Integer;
var Accept : Boolean;

  function DefaultHandler : Integer;
  begin
     Result := CallDefWndProc( Msg );
  end;

begin
   if (Msg.hwnd <> 0) and (fHandle = 0) then
      fHandle := Msg.hwnd;
   case Msg.message of
   WM_CLOSE: begin
                if Assigned( fOnClose ) then
                begin
                   Accept := True;
                   fOnClose( Self, Accept );
                   if (not Accept) and (not Applet.Terminated) then
                   begin
                      Result := 0;
                      Exit;
                   end;
                end;
                if Applet.MainWindow = Self then
                begin
                   Applet.Terminate;
                   PostQuitMessage( 0 );
                   Result := 0;
                   Exit;
                end;
             end;
   WM_NCDESTROY:  if Handle <> Applet.Handle then
             begin
                {
                PostMessage( Handle, CM_RELEASE, 0, 0 );
                }
                fDestroying := True;
                Free;
                Result := 0;
                Exit;
             end;
   {
   CM_RELEASE: begin
                Free;
                Result := 0;
                Exit;
             end;
   }
   WM_SIZE:  begin
                Result := DefaultHandler;
                case Msg.wParam of
                  SIZENORMAL: fWindowState := xwsNormal;
                  SIZEICONIC: fWindowState := xwsMinimized;
                  SIZEFULLSCREEN: fWindowState := xwsMaximized;
                end;
                Exit;
             end;
   WM_SHOWWINDOW:
             begin
                case Msg.lParam of
                  SW_PARENTCLOSING:
                     begin
                        if IsIconic( Handle ) then
                           fShowAction := SW_SHOWMINNOACTIVE
                        else
                        if IsZoomed( Handle ) then
                           fShowAction := SW_SHOWMAXIMIZED
                        else
                           fShowAction := SW_SHOWNOACTIVATE;
                     end;
                  SW_PARENTOPENING:
                     begin
                        if fShowAction <> 0 then
                        begin
                           ShowWindow( Handle, fShowAction );
                           fShowAction := 0;
                        end;
                        Result := 0;
                        Exit;
                     end;
                end;
             end;
   WM_SysCommand:
             begin
                if (Msg.wParam and $FFF0) = SC_MINIMIZE then
                if Applet.MainWindow = Self then
                begin
                   Applet.Minimize;
                   Result := 0;
                   Exit;
                end;
             end;
   WM_SETFOCUS:
             begin
                Applet.CurrentWindow := Self;
             end;
   WM_SETCURSOR:
             begin
                if LOWORD( Msg.lParam ) = HTCLIENT then
                begin
                   Windows.SetCursor( Cursor );
                   Result := 1;
                   Exit;
                end;
             end;

   end;
   Result := DefaultHandler;
end;

procedure XWindow.SetClsStyle( Value : Integer );
begin
   if fClsStyle = Value then Exit;
   fClsStyle := Value;
   if fHandle = 0 then Exit;
   SetClassLong( fHandle, GCL_STYLE, Value );
end;

procedure XWindow.SetStyle( Value : Integer );
begin
   if fStyle = Value then Exit;
   fStyle := Value;
   if fHandle = 0 then Exit;
   SetWindowLong( fHandle, GWL_STYLE, Value );
end;

procedure XWindow.SetExStyle( Value : Integer );
begin
   if fExStyle = Value then Exit;
   fExStyle := Value;
   if fHandle = 0 then Exit;
   SetWindowLong( fHandle, GWL_EXSTYLE, Value );
end;

procedure XWindow.SetCursor( Value : HCursor );
begin
   if fCursor = Value then Exit;
   fCursor := Value;
   if fHandle = 0 then Exit;
   SetClassLong( fHandle, GCL_HCURSOR, Value );
end;

procedure XWindow.SetIcon( Value : HIcon );
begin
   if fIcon = Value then Exit;
   fIcon := Value;
   if fHandle = 0 then Exit;
   SetClassLong( fHandle, GCL_HICON, Value );
   SendMessage( fHandle, WM_SETICON, 1, fIcon );
end;

procedure XWindow.SetMenu( Value : HMenu );
begin
   if fMenu = Value then Exit;
   if fMenu <> 0 then
      DestroyMenu( fMenu );
   fMenu := Value;
   if fHandle = 0 then Exit;
   //SetClassLong( fHandle, GCL_MENUNAME, Value );
   Windows.SetMenu( fHandle, fMenu );
end;

procedure XWindow.SetCaption( Value : String );
var L : Integer;
begin
   if (fCaption = nil) and (Value = '') or (fCaption = Value) then Exit;
   if fCaption <> nil then
   begin
      FreeMem( fCaption );
      fCaption := nil;
   end;
   if Value <> '' then
   begin
      L := Length( Value ) + 1;
      GetMem( fCaption, L );
      Move( Value[ 1 ], fCaption[ 0 ], L );
   end;
   if fHandle = 0 then Exit;
   SetWindowText( fHandle, FCaption );
end;

function XWindow.GetVisible : Boolean;
begin
   if fHandle <> 0 then
      inherited SetVisible( IsWindowVisible( fHandle ) );
   Result := inherited GetVisible;
end;

procedure XWindow.SetVisible( Value : Boolean );
begin
   if Visible = Value then Exit;
   inherited SetVisible( Value );
   if Value then
      fStyle := fStyle or WS_VISIBLE
   else
      fStyle := fStyle and not WS_VISIBLE;
   if fHandle = 0 then Exit;
   if Value then
      ShowWindow( fHandle, SW_SHOWNA )
   else
      ShowWindow( fHandle, SW_HIDE );
end;

procedure XWindow.SetColor( Value : XColor );
begin
   if Color = Value then Exit;
   inherited SetColor( Value );
   if fHandle = 0 then Exit;
   SetClassLong( fHandle, GCL_HBRBACKGROUND, Value );
end;

function XWindow.GetBoundsRect : TRect;
begin
   if Handle = 0 then
      Result := inherited GetBoundsRect
   else
      GetWindowRect( Handle, Result );
end;

procedure XWindow.SetBoundsRect( Value : TRect );
var Rect : TRect;
begin
   Rect := BoundsRect;
   if Value.Left = Rect.Left then
   if Value.Top = Rect.Top then
   if Value.Right = Rect.Right then
   if Value.Bottom = Rect.Bottom then Exit;
   inherited SetBoundsRect( Value );
   Rect := Value;
   if fHandle = 0 then Exit;
   SetWindowPos( fHandle, 0, Rect.Left, Rect.Top, Rect.Right - Rect.Left,
                 Rect.Bottom - Rect.Top, SWP_NOZORDER );
end;


{
function XWindow.GetWidth : Integer;
var R : TRect;
begin
   Result := BoundsRect.Right - BoundsRect.Left;
   if (fHandle <> 0) then
   begin
      GetWindowRect( fHandle, R );
      Result := R.Right - R.Left;
   end;
end;

procedure XWindow.SetWidth( Value : Integer );
var R : TRect;
begin
   R := BoundsRect;
   if Value = R.Right - R.Left then Exit;
   if R.Left = R.Right then
   if fHandle <> 0 then
      GetWindowRect( fHandle, R );
   R.Right := R.Left + Value;
   SetBoundsRect( R );
end;

function XWindow.GetHeight : Integer;
var R : TRect;
begin
   Result := BoundsRect.Bottom - BoundsRect.Top;
   if (fHandle <> 0) then
   begin
      GetWindowRect( fHandle, R );
      Result := R.Bottom - R.Top;
   end;
end;

procedure XWindow.SetHeight( Value : Integer );
var R : TRect;
begin
   R := BoundsRect;
   if Value = R.Bottom - R.Top then Exit;
   if R.Top = R.Bottom then
   if fHandle <> 0 then
      GetWindowRect( fHandle, R );
   R.Bottom := R.Top + Value;
   SetBoundsRect( R );
end;
}

procedure XWindow.SetWindowState( Value : XWindowState );
const
  ShowCommands: array[XWindowState] of Integer =
    (SW_SHOWNORMAL, SW_MINIMIZE, SW_SHOWMAXIMIZED);
begin
   if fWindowState <> Value then
   begin
      fWindowState := Value;
      ShowWindow(Handle, ShowCommands[Value]);
   end;
end;

procedure XWindow.Show;
begin
   if Handle = 0 then
      CreateWindow;
   Visible := True;
   SetForegroundWindow( Handle );
   SetFocus( Handle );
end;

procedure XWindow.Hide;
begin
   Visible := False;
end;

function XWindow.ClientToScreen( P : TPoint ) : TPoint;
begin
   Result := P;
   if Handle <> 0 then
      Windows.ClientToScreen( Handle, Result );
end;

function XWindow.ScreenToClient( P : TPoint ) : TPoint;
begin
   Result := P;
   if Handle <> 0 then
      Windows.ScreenToClient( Handle, Result );
end;

function XWindow.ClientRect : TRect;
begin
   if Handle <> 0 then
      GetClientRect( Handle, Result )
   else
      Result := XRect( 0, 0, Width - 8, Height - 20 );
end;

procedure XWindow.Invalidate;
begin
   if Handle = 0 then Exit;
   Windows.InvalidateRect( Handle, nil, True );
end;

function XWindow.GetIcon : HIcon;
begin
   Result := fIcon;
   if Result = 0 then
      Result := Applet.Icon;
end;

function XWindow.GetWindowHandle: HWnd;
begin
   if fHandle = 0 then
   begin
      if CreatingWindow = Self then
      begin
         Result := HWnd(-1);
         Exit;
      end;
      CreateWindow;
   end;
   Result := fHandle;
end;

function XWindow.GetHandle: HWnd;
begin
   Result := fHandle;
end;

procedure XWindow.CreateParams(var Params: TCreateParams);
var R : TRect;
begin
   FillChar( Params, Sizeof( Params ), 0 );
   Params.Caption := FCaption;
   Params.Style := FStyle;
   if not Enabled then
      Params.Style := Params.Style or WS_DISABLED;
   Params.ExStyle := FExStyle;
   Params.WindowClass.style := FClsStyle;
   Params.WindowClass.lpfnWndProc := @DefWindowProc;
   Params.WindowClass.hCursor := LoadCursor( 0, IDC_ARROW );
   Params.WindowClass.hInstance := hInstance;
   StrCopy( Params.WinClassName, PChar( String( ClassName ) ) );
   Params.WndParent := GetParentWnd( True );
   R := BoundsRect;
   R.Right := R.Right - R.Left;
   R.Bottom := R.Bottom - R.Top;
   if Width = 0 then
      R.Right := Integer( CW_UseDefault );
   if Height = 0 then
      R.Bottom := Integer( CW_UseDefault );
   if (Width = 0) and (Height = 0) and (Left = 0) and (Top = 0) then
      R.TopLeft := XPoint( Integer( CW_UseDefault ), Integer( CW_UseDefault ) );
   Params.X := R.Left;
   Params.Y := R.Top;
   Params.Width := R.Right;
   Params.Height := R.Bottom;
end;

procedure XWindow.CreateSubclass(var Params: TCreateParams;
  ControlClassName: PChar);
const
  CS_OFF = CS_OWNDC or CS_CLASSDC or CS_PARENTDC or CS_GLOBALCLASS;
  CS_ON = CS_VREDRAW or CS_HREDRAW;
var
  SaveInstance: THandle;
begin
  if ControlClassName <> nil then
    with Params do
    begin
      SaveInstance := WindowClass.hInstance;
      if not GetClassInfo(HInstance, ControlClassName, WindowClass) and
        not GetClassInfo(0, ControlClassName, WindowClass) and
        not GetClassInfo(MainInstance, ControlClassName, WindowClass) then
        GetClassInfo(WindowClass.hInstance, ControlClassName, WindowClass);
      WindowClass.hInstance := SaveInstance;
      WindowClass.style := WindowClass.style and not CS_OFF or CS_ON;
    end;
end;

function XWindow.GetCaption: String;
begin
   Result := FCaption;
end;

function XWindow.CallDefWndProc(var Msg: TMsg): Integer;
begin
   Result := CallWindowProc( FDefWndProc, FHandle, Msg.message,
                             Msg.wParam, Msg.lParam );
end;

initialization

   GlobalWindowsManager := XWindowsManager.Create;

finalization

   GlobalWindowsManager.Free;

end.
